# Copyright (C) 2018-2020 The python-litecoin-utils developers
#
# This file is part of python-litecoin-utils
#
# It is subject to the license terms in the LICENSE file found in the top-level
# directory of this distribution.
#
# No part of python-litecoin-utils, including this file, may be copied, modified,
# propagated, or distributed except according to the terms contained in the
# LICENSE file.

import hashlib
import re
import struct
from abc import ABC, abstractmethod
from base64 import b64decode, b64encode
from binascii import hexlify, unhexlify

from base58check import b58decode, b58encode
from ecdsa import SECP256k1, SigningKey, VerifyingKey, ellipticcurve, numbertheory
from ecdsa.util import sigdecode_string, sigencode_der, sigencode_string
from sympy.ntheory import sqrt_mod

import litecoinutils.bech32
import litecoinutils.script
from litecoinutils.constants import (
    NETWORK_P2PKH_PREFIXES,
    NETWORK_P2SH_PREFIXES,
    NETWORK_SEGWIT_PREFIXES,
    NETWORK_WIF_PREFIXES,
    P2PKH_ADDRESS,
    P2SH_ADDRESS,
    P2WPKH_ADDRESS_V0,
    P2WSH_ADDRESS_V0,
    SIGHASH_ALL,
)
from litecoinutils.setup import get_network
from litecoinutils.utils import encode_varint

# ECDSA curve using secp256k1 is defined by: y**2 = x**3 + 7
# This is done modulo p which (secp256k1) is:
# p is the finite field prime number and is equal to:
# 2^256 - 2^32 - 2^9 - 2^8 - 2^7 - 2^6 - 2^4 - 1
# Note that we could also get that from ecdsa lib from the curve, e.g.:
# SECP256k1.__dict__['curve'].__dict__['_CurveFp__p']
_p = 0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFEFFFFFC2F
# Curve's a and b are (y**2 = x**3 + a*x + b)
_a = 0x0000000000000000000000000000000000000000000000000000000000000000
_b = 0x0000000000000000000000000000000000000000000000000000000000000007
# Curve's generator point is:
_Gx = 0x79BE667EF9DCBBAC55A06295CE870B07029BFCDB2DCE28D959F2815B16F81798
_Gy = 0x483ada7726a3c4655da4fbfc0e1108a8fd17b448a68554199c47d08ffb10d4b8
# prime number of points in the group (the order)
_order = 0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFEBAAEDCE6AF48A03BBFD25E8CD0364141

# The ECDSA curve (secp256k1) is:
# Note that we could get that from ecdsa lib, e.g.:
# SECP256k1.__dict__['curve']
_curve = ellipticcurve.CurveFp( _p, _a, _b )

# The generator base point is:
# Note that we could get that from ecdsa lib, e.g.:
# SECP256k1.__dict__['generator']
_G = ellipticcurve.Point( _curve, _Gx, _Gy, _order )




# method used by both PrivateKey and PublicKey - TODO clean - add in another module?
def add_magic_prefix(message):
    magic_prefix = b"\x19Litecoin Signed Message:\n"
    message_size = encode_varint(len(message))
    message_encoded = message.encode("utf-8")
    return magic_prefix + message_size + message_encoded



class PrivateKey:
    """Represents an ECDSA private key.

    Attributes
    ----------
    key : bytes
        the raw key of 32 bytes

    Methods
    -------
    from_wif(wif)
        creates an object from a WIF of WIFC format (string)
    to_wif(compressed=True)
        returns as WIFC (compressed) or WIF format (string)
    to_bytes()
        returns the key's raw bytes
    sign_message(message, compressed=True)
        signs the message's digest and returns the signature
    sign_transaction(tx, compressed=True)
        signs the transaction's digest and returns the signature
    get_public_key()
        returns the corresponding PublicKey object
    """

    def __init__(self, wif=None, secret_exponent=None):
        """With no parameters a random key is created

        Parameters
        ----------
        wif : str, optional
            the key in WIF of WIFC format (default None)
        secret_exponent : int, optional
            used to create a specific key deterministically (default None)
        """

        if not secret_exponent and not wif:
            self.key = SigningKey.generate(curve=SECP256k1)
        else:
            if wif:
                self._from_wif(wif)
            elif secret_exponent:
                self.key = SigningKey.from_secret_exponent(secret_exponent,
                                                           curve=SECP256k1)

    def to_bytes(self):
        """Returns key's bytes"""

        return self.key.to_string()


    @classmethod
    def from_wif(cls, wif):
        """Creates key from WIFC or WIF format key"""

        return cls(wif=wif)


    # expects wif in hex string
    def _from_wif(self, wif):
        """Creates key from WIFC or WIF format key

        Check to_wif for the detailed process. From WIF is the reverse.

        Raises
        ------
        ValueError
            if the checksum is wrong or if the WIF/WIFC is not from the
            configured network.
        """

        wif_utf = wif.encode('utf-8')

        # decode base58check get key bytes plus checksum
        data_bytes = b58decode( wif_utf )
        key_bytes = data_bytes[:-4]
        checksum = data_bytes[-4:]

        # verify key with checksum
        data_hash = hashlib.sha256(hashlib.sha256(key_bytes).digest()).digest()
        if not checksum == data_hash[0:4]:
            raise ValueError('Checksum is wrong. Possible mistype?')

        # get network prefix and check with current setup
        network_prefix = key_bytes[:1]
        if NETWORK_WIF_PREFIXES[get_network()] != network_prefix:
            raise ValueError('Using the wrong network!')

        # remove network prefix
        key_bytes = key_bytes[1:]

        # check length of bytes and if > 32 then compressed
        # use this to instantite an ecdsa key
        if len(key_bytes) > 32:
            self.key = SigningKey.from_string(key_bytes[:-1], curve=SECP256k1)
        else:
            self.key = SigningKey.from_string(key_bytes, curve=SECP256k1)


    def to_wif(self, compressed=True):
        """Returns key in WIFC or WIF string

        |  Pseudocode:
        |      network_prefix = (1 byte version number)
        |      data = network_prefix + (32 bytes number/key) [ + 0x01 if compressed ]
        |      data_hash = SHA-256( SHA-256( data ) )
        |      checksum = (first 4 bytes of data_hash)
        |      wif = Base58CheckEncode( data + checksum )
        """

        # add network prefix to the key
        data = NETWORK_WIF_PREFIXES[get_network()] + self.to_bytes()

        if compressed == True:
            data += b'\x01'

        # double hash and get the first 4 bytes for checksum
        data_hash = hashlib.sha256(hashlib.sha256(data).digest()).digest()
        checksum = data_hash[0:4]

        # suffix the key bytes with the checksum and encode to base58check
        wif = b58encode( data + checksum )

        return wif.decode('utf-8')


    def sign_message(self, message, compressed=True):
        """Signs the message with the private key (deterministically)

        Bitcoin uses a compact format for message signatures (for tx sigs it
        uses normal DER format). The format has the normal r and s parameters
        that ECDSA signatures have but also includes a prefix which encodes
        extra information. Using the prefix the public key can be
        reconstructed when verifying the signature.

        |  Prefix values:
        |      27 - 0x1B = first key with even y
        |      28 - 0x1C = first key with odd y
        |      29 - 0x1D = second key with even y
        |      30 - 0x1E = second key with odd y

        If key is compressed add 4 (31 - 0x1F, 32 - 0x20, 33 - 0x21, 34 - 0x22 respectively)

        Returns a Bitcoin compact signature in Base64
        """

        # All litecoin signatures include the magic prefix. It is just a string
        # added to the message to distinguish Litecoin-specific messages.
        message_magic = add_magic_prefix(message)

        # create message digest -- note double hashing
        message_digest = hashlib.sha256( hashlib.sha256(message_magic).digest() ).digest()

        #
        # sign non-deterministically - no reason
        #signature = self.key.sign_digest(message_digest,
        #                                 sigencode=sigencode_string)

        # deterministic signing
        signature = self.key.sign_digest_deterministic(message_digest,
                                                       sigencode=sigencode_string,
                                                       hashfunc=hashlib.sha256)
        prefix = 27
        if compressed:
            prefix += 4

        address = self.get_public_key().get_address(compressed=compressed).to_string()
        for i in range(prefix, prefix + 4):
            recid = chr(i).encode('utf-8')
            sig = b64encode( recid + signature ).decode('utf-8')
            try:
                if PublicKey.verify_message(address, sig, message):
                    return sig
            except:
                continue


    def sign_input(self, tx, txin_index, script, sighash=SIGHASH_ALL):
        # the tx knows how to calculate the digest for the corresponding
        # sighash)
        tx_digest = tx.get_transaction_digest(txin_index, script, sighash)
        return self._sign_input(tx_digest, sighash)


    def sign_segwit_input(self, tx, txin_index, script, amount, sighash=SIGHASH_ALL):
        # the tx knows how to calculate the digest for the corresponding
        # sighash)
        tx_digest = tx.get_transaction_segwit_digest(txin_index, script, amount, sighash)
        return self._sign_input(tx_digest, sighash)


    def _sign_input(self, tx_digest, sighash=SIGHASH_ALL):
        """Signs a transaction input with the private key

        Bitcoin uses the normal DER format for transactions. Each input is
        signed separately (thus txin_index is required). The script of the
        input we wish to spend is required and replaces the transaction's
        script sig in order to calculate the correct transaction hash (which
        is what is actually signed!)

        Returns a signature for that input
        """

        # note that deterministic signing is used
        signature = self.key.sign_digest_deterministic(tx_digest,
                                                       sigencode=sigencode_der,
                                                       hashfunc=hashlib.sha256)

        # make sure that signature complies with Low S standardness rule of
        # BIP62: https://github.com/litecoin/bips/blob/master/bip-0062.mediawiki
        #
        # Both R ans S cannot start with 0x00 (be signed as negative) unless
        # they are higher than 2^128 or start with 0x80.
        #
        # The S part of the signature is equivalent to (order-S). This allows
        # for txid malleability attacks where S is modified with (order-S) and
        # thus a valid signature... but the txid hash would be different!
        #
        # For this reason Low S standardness rule specifies that all S's need
        # to be less than half of the curve order (SECP256k1). If it is not we
        # ensure it is by substrituting it with (order-S).

        # get DER values individually -- DER structure is:
        #   1-byte   -- 0x30 to specify a DER compound object (R,S)
        #   1-byte   -- length of the compound object
        #   1-byte   -- 0x02 to specify integer type for R
        #   1-byte   -- length of signature's R value
        #   variable -- R value
        #   1-byte   -- 0x02 to specify integer type for S
        #   1-byte   -- length of signature's S value
        #   variable -- S value

        der_prefix = signature[0]
        length_total = signature[1]
        der_type_int = signature[2]
        length_r = signature[3]
        R = signature[4:4+length_r]
        length_s = signature[5 + length_r]
        S = signature[5 + length_r + 1:]
        S_as_bigint = int( hexlify(S).decode('utf-8'), 16 )

        # update R, S if necessary -- in Bitcoin DER signatures' R should have a
        # prefix of 0x00 only if it starts with 0x80 or higher -- this was
        # implemented in Bitcoin Core of v0.17 to always be the case (however,
        # signatures are still valid even without a Low R value. Because R is
        # not mutable in the same way that S is, a low R value can only be
        # found by trying different nonves (RFC6979 - deterministic nonce
        # generation).
        # TODO to be 100% compliant with Bitcoin Core (still valid without it)

        # update S if necessary -- Low S standardness rule
        half_order = _order // 2
        # if S is larger than half the order then substructed from order and
        # use that as S since it is equivalent.
        if S_as_bigint > half_order:
            # make sure length is 33 bytes (it should be)
            assert length_s == 0x21

            new_S_as_bigint = _order - S_as_bigint
            # convert bigint to bytes
            new_S = unhexlify( format(new_S_as_bigint, 'x').zfill(64) )
            # new value should be 32 bytes
            assert len(new_S) == 0x20
            # reduce appropriate lengths
            length_s -= 1
            length_total -= 1
        else:
            new_S = S

        # reconstruct signature
        signature = struct.pack('BBBB', der_prefix, length_total, der_type_int, length_r) + R + \
                        struct.pack('BB', der_type_int, length_s) + new_S

        # add sighash in the signature -- as one byte!
        signature += struct.pack('B', sighash)

        # note that this is the final sig that needs to be added in the
        # script_sig (i.e. the DER signature plus the sighash)
        return hexlify(signature).decode('utf-8')


    def get_public_key(self):
        """Returns the corresponding PublicKey"""

        verifying_key = hexlify(self.key.get_verifying_key().to_string())
        return PublicKey( '04' + verifying_key.decode('utf-8') )


class PublicKey:
    """Represents an ECDSA public key.

    Attributes
    ----------
    key : bytes
        the raw public key of 64 bytes (x, y coordinates of the ECDSA curve)

    Methods
    -------
    from_hex(hex_str)
        creates an object from a hex string in SEC format
    from_message_signature(signature)
        NO-OP!
    verify_message(address, signature, message)
        Class method that constructs the public key, confirms the address and
        verifies the signature
    verify(signature, message)
        returns true if the message was signed with this public key's
        corresponding private key.
    to_hex(compressed=True)
        returns the key as hex string (in SEC format - compressed by default)
    to_bytes()
        returns the key's raw bytes
    to_hash160()
        returns the hash160 hex string of the public key
    get_address(compressed=True))
        returns the corresponding P2pkhAddress object
    get_segwit_address()
        returns the corresponding P2wpkhAddress object
    """


    def __init__(self, hex_str):
        """
        Parameters
        ----------
        hex_str : str
            the public key in hex string

        Raises
        ------
        TypeError
            If first byte of public key (corresponding to SEC format) is
            invalid.
        """

        # expects key as hex string - SEC format
        first_byte_in_hex = hex_str[:2] # 2 since a byte is represented by 2 hex characters
        hex_bytes = unhexlify(hex_str)

        # check if compressed or not
        if len(hex_bytes) > 33:
            # uncompressed - SEC format: 0x04 + x + y coordinates (x,y are 32 byte numbers)
            # remove first byte and instantiate ecdsa key
            self.key = VerifyingKey.from_string(hex_bytes[1:], curve=SECP256k1)
        else:
            # compressed - SEC FORMAT: 0x02|0x03 + x coordinate (if 02 then y
            # is even else y is odd. Calculate y and then instantiate the ecdsa key
            x_coord = int( hex_str[2:], 16 )

            # y = modulo_square_root( (x**3 + 7) mod p ) -- there will be 2 y values
            y_values = sqrt_mod( (x_coord**3 + 7) % _p, _p, True )

            # check SEC format's first byte to determine which of the 2 values to use
            if first_byte_in_hex == '02':
                # y is the even value
                if y_values[0] % 2 == 0:
                    y_coord = y_values[0]
                else:
                    y_coord = y_values[1]
            elif first_byte_in_hex == '03':
                # y is the odd value
                if y_values[0] % 2 == 0:
                    y_coord = y_values[1]
                else:
                    y_coord = y_values[0]
            else:
                raise TypeError("Invalid SEC compressed format")

            uncompressed_hex = "%0.64X%0.64X" % (x_coord, y_coord)
            uncompressed_hex_bytes = unhexlify(uncompressed_hex)
            self.key = VerifyingKey.from_string(uncompressed_hex_bytes, curve=SECP256k1)


    @classmethod
    def from_hex(cls, hex_str):
        """Creates a public key from a hex string (SEC format)"""

        return cls(hex_str)


    def to_bytes(self):
        """Returns key's bytes"""

        return self.key.to_string()


    def to_hex(self, compressed=True):
        """Returns public key as a hex string (SEC format - compressed by
        default)"""

        key_hex = hexlify(self.key.to_string())

        if compressed:
            # check if y is even or odd (02 even, 03 odd)
            if int(key_hex[-2:], 16) % 2 == 0:
                key_str = b'02' + key_hex[:64]
            else:
                key_str = b'03' + key_hex[:64]
        else:
            # uncompressed starts with 04
            key_str = b'04' + key_hex

        return key_str.decode('utf-8')


    @classmethod
    def from_message_signature(self, signature):
        # TODO implement (add signature=None in __init__, etc.)
        # TODO plus does this apply to DER signatures as well?
        #return cls(signature=signature)
        raise BaseException('NO-OP!')


    @classmethod
    def verify_message(self, address, signature, message):
        """Creates a public key from a message signature and verifies message

        Bitcoin uses a compact format for message signatures (for tx sigs it
        uses normal DER format). The format has the normal r and s parameters
        that ECDSA signatures have but also includes a prefix which encodes
        extra information. Using the prefix the public key can be
        reconstructed from the signature.

        |  Prefix values:
        |      27 - 0x1B = first key with even y
        |      28 - 0x1C = first key with odd y
        |      29 - 0x1D = second key with even y
        |      30 - 0x1E = second key with odd y

        If key is compressed add 4 (31 - 0x1F, 32 - 0x20, 33 - 0x21, 34 - 0x22 respectively)

        Raises
        ------
        ValueError
            If signature is invalid
        """

        sig = b64decode( signature.encode('utf-8') )
        if len(sig) != 65:
            raise ValueError('Invalid signature size')

        # get signature prefix, compressed and recid (which key is odd/even)
        prefix = sig[0]
        if prefix < 27 or prefix > 35:
            return False
        if prefix >= 31:
            compressed = True
            recid = prefix - 31
        else:
            compressed = False
            recid = prefix - 27

        # create message digest -- note double hashing
        message_magic = add_magic_prefix(message)
        message_digest = hashlib.sha256( hashlib.sha256(message_magic).digest() ).digest()

        #
        # use recid, r and s to get the point in the curve
        #

        # get signature's r and s
        r,s = sigdecode_string(sig[1:], _order)

        # ger R's x coordinate
        x = r + (recid // 2) * _order

        # get R's y coordinate (y**2 = x**3 + 7)
        y_values = sqrt_mod( (x**3 + 7) % _p, _p, True )
        if (y_values[0] - recid) % 2 == 0:
            y = y_values[0]
        else:
            y = y_values[1]

        # get R (recovered ephemeral key) from x,y
        R = ellipticcurve.Point(_curve, x, y, _order)

        # get e (hash of message encoded as big integer)
        e = int(hexlify(message_digest), 16)

        # compute public key Q = r^-1 (sR - eG)
        # because Point substraction is not defined we will instead use:
        # Q = r^-1 (sR + (-eG) )
        minus_e = -e % _order
        inv_r = numbertheory.inverse_mod(r, _order)
        Q = inv_r * ( s*R + minus_e*_G )

        # instantiate the public key and verify message
        public_key = VerifyingKey.from_public_point( Q, curve = SECP256k1 )
        key_hex = hexlify(public_key.to_string()).decode('utf-8')
        pubkey = PublicKey.from_hex('04' + key_hex)
        if not pubkey.verify(signature, message):
            return False

        # confirm that the address provided corresponds to that public key
        if pubkey.get_address(compressed=compressed).to_string() != address:
            return False

        return True


    def verify(self, signature, message):
        """Verifies that the message was signed with this public key's
        corresponding private key."""

        # All litecoin signatures include the magic prefix. It is just a string
        # added to the message to distinguish Litecoin-specific messages.
        message_magic = add_magic_prefix(message)

        # create message digest -- note double hashing
        message_digest = hashlib.sha256( hashlib.sha256(message_magic).digest()).digest()

        signature_bytes = b64decode( signature.encode('utf-8') )

        # verify -- ignore first byte of compact signature
        return self.key.verify_digest(signature_bytes[1:],
                                      message_digest,
                                      sigdecode=sigdecode_string)


    def _to_hash160(self, compressed=True):
        """Returns the RIPEMD( SHA256( ) ) of the public key in bytes"""

        pubkey = unhexlify( self.to_hex(compressed) )
        hashsha256 = hashlib.sha256(pubkey).digest()
        hashripemd160 = hashlib.new('ripemd160')
        hashripemd160.update(hashsha256)
        hash160 = hashripemd160.digest()
        return hash160

    def to_hash160(self, compressed=True):
        """Returns the RIPEMD( SHA256( ) ) of the public key in hex"""

        return hexlify(self._to_hash160(compressed)).decode('utf-8')


    def get_address(self, compressed=True):
        """Returns the corresponding P2PKH Address (default compressed)"""

        hash160 = self._to_hash160(compressed)
        addr_string_hex = hexlify(hash160).decode('utf-8')
        return P2pkhAddress(hash160=addr_string_hex)


    def get_segwit_address(self):
        """Returns the corresponding P2WPKH address

        Only compressed is allowed. It is otherwise identical to normal P2PKH
        address.
        """
        hash160 = self._to_hash160(True)
        addr_string_hex = hexlify(hash160).decode('utf-8')
        return P2wpkhAddress(witness_hash=addr_string_hex)



class Address(ABC):
    """Represents a Bitcoin address

    Attributes
    ----------
    hash160 : str
        the hash160 string representation of the address; hash160 represents
        two consequtive hashes of the public key or the redeam script, first
        a SHA-256 and then an RIPEMD-160

    Methods
    -------
    from_address(address)
        instantiates an object from address string encoding
    from_hash160(hash160_str)
        instantiates an object from a hash160 hex string
    from_script(redeem_script)
        instantiates an object from a redeem_script
    to_string()
        returns the address's string encoding
    to_hash160()
        returns the address's hash160 hex string representation

    Raises
    ------
    TypeError
        No parameters passed
    ValueError
        If an invalid address or hash160 is provided.
    """
    @abstractmethod
    def __init__(self, address=None, hash160=None, script=None):
        """
        Parameters
        ----------
        address : str
            the address as a string
        hash160 : str
            the hash160 hex string representation
        script : Script object
            instantiates an Address object from a redeem script

        Raises
        ------
        TypeError
            No parameters passed
        ValueError
            If an invalid address or hash160 is provided.
        """

        if hash160:
            if self._is_hash160_valid(hash160):
                self.hash160 = hash160
            else:
                raise ValueError("Invalid value for parameter hash160.")
        elif address:
            if self._is_address_valid(address):
                self.hash160 = self._address_to_hash160(address)
            else:
                raise ValueError("Invalid value for parameter address.")
        elif script:
            # TODO for now just check that is an instance of Script
            if isinstance(script, litecoinutils.script.Script):
                self.hash160 = self._script_to_hash160(script)
            else:
                raise TypeError("A Script class is required.")
        else:
            raise TypeError("A valid address or hash160 is required.")


    @classmethod
    def from_address(cls, address):
        """Creates and address object from an address string"""

        return cls(address=address)


    @classmethod
    def from_hash160(cls, hash160):
        """Creates and address object from a hash160 string"""

        return cls(hash160=hash160)


    @classmethod
    def from_script(cls, script):
        """Creates and address object from a Script object"""

        return cls(script=script)


    def _address_to_hash160(self, address):
        """Converts an address to it's hash160 equivalent

	Base58CheckDecode the address and remove network_prefix and checksum.
	"""

        addr_encoded = address.encode('utf-8')
        data_checksum = b58decode( addr_encoded )
        network_prefix = data_checksum[:1]
        data = data_checksum[1:-4]
        #checksum = data_checksum[-4:]
        return hexlify(data).decode('utf-8')


    def _script_to_hash160(self, script):
        """Converts a script to it's hash160 equivalent

        RIPEMD160( SHA256( script ) ) - required for P2SH addresses
	"""

        script_bytes = script.to_bytes()
        hashsha256 = hashlib.sha256(script_bytes).digest()
        hashripemd160 = hashlib.new('ripemd160')
        hashripemd160.update(hashsha256)
        hash160 = hashripemd160.digest()
        return hexlify(hash160).decode('utf-8')


    def _is_hash160_valid(self, hash160):
        """Checks is a hash160 hex string is valid"""

        # check the size -- should be 20 bytes, 40 characters in hexadecimal string
        if len(hash160) != 40:
            return False

        # check all (string) digits are hex
        try:
            int(hash160, 16)
            return True
        except ValueError:
            return False


    def _is_address_valid(self, address):
        """Checks is an address string is valid"""

        digits_58_pattern = r'[^123456789ABCDEFGHJKLMNPQRSTUVWXYZabcdefghijkmnopqrstuvwxyz]'

        # check for invalid characters
        if re.search(digits_58_pattern, address):
            return False

        # check for length (26-35 characters)
        # TODO: need to confirm the possible length!
        if len(address) < 26 or len(address) > 35:
            return False

        # get data, network_prefix and checksum
        data_checksum = b58decode( address.encode('utf-8') )
        data = data_checksum[:-4]
        network_prefix = data_checksum[:1]
        checksum = data_checksum[-4:]

        # check correct network (depending on address type)
        if self.get_type() == P2PKH_ADDRESS:
            if network_prefix != NETWORK_P2PKH_PREFIXES[get_network()]:
                return False
        elif self.get_type() == P2SH_ADDRESS:
            if network_prefix != NETWORK_P2SH_PREFIXES[get_network()]:
                return False

        # check address' checksum
        data_hash = hashlib.sha256(hashlib.sha256(data).digest()).digest()

        if data_hash[0:4] != checksum:
            return False

        return True


    def to_hash160(self):
        """Returns as hash160 hex string"""

        return self.hash160


    def to_string(self):
        """Returns as address string

        |  Pseudocode:
        |      network_prefix = (1 byte version number)
        |      data = network_prefix + hash160_bytes
        |      data_hash = SHA-256( SHA-256( hash160_bytes ) )
        |      checksum = (first 4 bytes of data_hash)
        |      address_bytes = Base58CheckEncode( data + checksum )
        """
        hash160_encoded = self.hash160.encode('utf-8')
        hash160_bytes = unhexlify(hash160_encoded)

        if self.get_type() == P2PKH_ADDRESS:
            data = NETWORK_P2PKH_PREFIXES[get_network()] + hash160_bytes
        elif self.get_type() == P2SH_ADDRESS:
            data = NETWORK_P2SH_PREFIXES[get_network()] + hash160_bytes

        data_hash = hashlib.sha256(hashlib.sha256(data).digest()).digest()
        checksum = data_hash[0:4]
        address_bytes = b58encode( data + checksum )

        return address_bytes.decode('utf-8')


class P2pkhAddress(Address):
    """Encapsulates a P2PKH address.

    Check Address class for details

    Methods
    -------
    to_script_pub_key()
        returns the scriptPubKey (P2PKH) that corresponds to this address
    get_type()
        returns the type of address
    """

    def __init__(self, address=None, hash160=None):
        super().__init__(address=address, hash160=hash160)

    def to_script_pub_key(self):
        """Returns the scriptPubKey (P2PKH) that corresponds to this address"""
        return litecoinutils.script.Script(['OP_DUP', 'OP_HASH160',
                                           self.to_hash160(), 'OP_EQUALVERIFY',
                                           'OP_CHECKSIG'])

    def get_type(self):
        """Returns the type of address"""
        return P2PKH_ADDRESS


class P2shAddress(Address):
    """Encapsulates a P2SH address.

    Check Address class for details

    Methods
    -------
    get_type()
        returns the type of address
    """

    def __init__(self, address=None, hash160=None, script=None):
        super().__init__(address=address, hash160=hash160, script=script)

    def get_type(self):
        """Returns the type of address"""
        return P2SH_ADDRESS





class SegwitAddress(ABC):
    """Represents a Bitcoin segwit address

    Note that currently the python bech32 reference implementation is used (by
    Pieter Wuille).

    Attributes
    ----------
    witness_hash : str
        the hash string representation of either the address; it can be either
        a public key hash (P2WPKH) or the hash of the script (P2WSH)

    Methods
    -------
    from_address(address)
        instantiates an object from address string encoding
    from_hash(hash_str)
        instantiates an object from a hash hex string
    from_script(witness_script)
        instantiates an object from a witness_script
    to_string()
        returns the address's string encoding (Bech32)
    to_hash()
        returns the address's hash hex string representation

    Raises
    ------
    TypeError
        No parameters passed
    ValueError
        If an invalid address or hash is provided.
    """
    @abstractmethod
    def __init__(self, address=None, witness_hash=None, script=None,
                 version=P2WPKH_ADDRESS_V0):
        """
        Parameters
        ----------
        address : str
            the address as a string
        witness_hash : str
            the hash hex string representation
        script : Script object
            instantiates an Address object from a witness script
        version : str
            specifies the default segwit version

        Raises
        ------
        TypeError
            No parameters passed
        ValueError
            If an invalid address or hash is provided.
        """

        self.version = version
        if self.version == P2WPKH_ADDRESS_V0 or self.version == P2WSH_ADDRESS_V0:
            self.segwit_num_version = 0

        if witness_hash:
            self.witness_hash = witness_hash
        elif address:
            self.witness_hash = self._address_to_hash(address)
        elif script:
            # TODO for now just check that is an instance of Script
            if isinstance(script, litecoinutils.script.Script):
                self.witness_hash = self._script_to_hash(script)
            else:
                raise TypeError("A Script class is required.")
        else:
            raise TypeError("A valid address or hash is required.")


    @classmethod
    def from_address(cls, address):
        """Creates and address object from an address string"""

        return cls(address=address)


    @classmethod
    def from_hash(cls, witness_hash):
        """Creates and address object from a hash string"""

        return cls(witness_hash=witness_hash)


    @classmethod
    def from_script(cls, script):
        """Creates and address object from a Script object"""

        return cls(script=script)


    def _address_to_hash(self, address):
        """Converts an address to it's hash equivalent

	The size of the address determines between P2WPKH and P2WSH.
        Then Bech32 decodes the address removing network prefix, checksum,
        witness version.

        Uses a segwit's python reference implementation for now. (TODO)
	"""

        witness_version, witness_int_array = litecoinutils.bech32.decode(NETWORK_SEGWIT_PREFIXES[get_network()], address)
        if witness_version == None:
            raise ValueError("Invalid value for parameter address.")
        if witness_version != self.segwit_num_version:
            raise TypeError("Invalid segwit version.")

        return hexlify( bytes(witness_int_array) ).decode('utf-8')


    def _script_to_hash(self, script):
        """Converts a script to it's hash equivalent

        Uses a segwit's python reference implementation for now. (TODO)
	"""

        script_bytes = script.to_bytes()
        hashsha256 = hashlib.sha256(script_bytes).digest()
        return hexlify(hashsha256).decode('utf-8')


    def to_hash(self):
        """Returns as hash hex string"""

        return self.witness_hash


    def to_string(self):
        """Returns as address string

        Uses a segwit's python reference implementation for now. (TODO)
        """

        # convert hex string hash to int array (required by bech32 lib)
        hash_bytes = unhexlify( self.witness_hash.encode('utf-8') )
        witness_int_array = memoryview(hash_bytes).tolist()

        return litecoinutils.bech32.encode(NETWORK_SEGWIT_PREFIXES[get_network()],
                                          self.segwit_num_version, witness_int_array)



class P2wpkhAddress(SegwitAddress):
    """Encapsulates a P2WPKH address.

    Check Address class for details

    Methods
    -------
    to_script_pub_key()
        returns the scriptPubKey of a P2WPKH witness script
    get_type()
        returns the type of address
    """

    # TODO allow creation directly from Bech32 address !!!!!!
    def __init__(self, address=None, witness_hash=None,
                 version=P2WPKH_ADDRESS_V0):
        """Allow creation only from hash160 of public key"""

        super().__init__(address=address, witness_hash=witness_hash,
                         version=version)


    def to_script_pub_key(self):
        """Returns the scriptPubKey of a P2WPKH witness script"""
        return litecoinutils.script.Script(['OP_0', self.to_hash()])


    def get_type(self):
        """Returns the type of address"""
        return self.version


class P2wshAddress(SegwitAddress):
    """Encapsulates a P2WSH address.

    Check Address class for details

    Methods
    -------
    from_script(witness_script)
        instantiates an object from a witness_script
    get_type()
        returns the type of address
    """

    def __init__(self, address=None, witness_hash=None, script=None,
                 version=P2WSH_ADDRESS_V0):
        """Allow creation only from hash160 of public key"""

        super().__init__(address=None, witness_hash=None, script=script,
                         version=version)


    def to_script_pub_key(self):
        """Returns the scriptPubKey of a P2WPKH witness script"""
        return litecoinutils.script.Script(['OP_0', self.to_hash()])


    def get_type(self):
        """Returns the type of address"""
        return self.version



def main():
    pass

if __name__ == "__main__":
    main()



# Copyright (C) 2018-2020 The python-litecoin-utils developers
#
# This file is part of python-litecoin-utils
#
# It is subject to the license terms in the LICENSE file found in the top-level
# directory of this distribution.
#
# No part of python-litecoin-utils, including this file, may be copied, modified,
# propagated, or distributed except according to the terms contained in the
# LICENSE file.

NETWORK = None

networks = {'mainnet', 'testnet', 'regtest'}

def setup(network='mainnet'):
    global NETWORK
    NETWORK = network
    return NETWORK


def get_network():
    global NETWORK
    return NETWORK


def is_mainnet():
    global NETWORK
    if NETWORK == 'mainnet':
        return True
    else:
        return False

def is_testnet():
    global NETWORK
    if NETWORK == 'testnet':
        return True
    else:
        return False

def is_regtest():
    global NETWORK
    if NETWORK == 'regtest':
        return True
    else:
        return False



# Copyright (C) 2018-2020 The python-litecoin-utils developers
#
# This file is part of python-litecoin-utils
#
# It is subject to the license terms in the LICENSE file found in the top-level
# directory of this distribution.
#
# No part of python-litecoin-utils, including this file, may be copied, modified,
# propagated, or distributed except according to the terms contained in the
# LICENSE file.

from bitcoinrpc.authproxy import AuthServiceProxy

from litecoinutils.setup import get_network
from litecoinutils.constants import NETWORK_DEFAULT_PORTS


class NodeProxy:
    """Simple Bitcoin node proxy that can call all of Bitcoin's JSON-RPC functionality.

    Attributes
    ----------
    proxy : object
        a litecoinrpc AuthServiceProxy object
    """

    def __init__(self, rpcuser=None, rpcpassword=None, host=None, port=None):
        """Connects to node using credentials given

        Parameters
        ----------
        rpcuser : str
            as defined in litecoin.conf
        rpcpassword : str
            as defined in litecoin.conf
        host : str, optional
            host where the Bitcoin node resides; defaults to 127.0.0.1
        port : int, optional
            port to connect to; uses default ports according to network

        Raises
        ------
        ValueError
            if rpcuser and/or rpcpassword are not specified
        """

        if not rpcuser or not rpcpassword:
            raise ValueError('rpcuser or rpcpassword is missing')

        if not host:
            host = '127.0.0.1'

        if not port:
            port = NETWORK_DEFAULT_PORTS[get_network()]

        self.proxy = AuthServiceProxy("http://{}:{}@{}:{}".format(rpcuser, rpcpassword, host, port))


    def get_proxy(self):
        """Returns litecoinrpc AuthServiceProxy object"""
        return self.proxy




# Copyright (C) 2018-2020 The python-litecoin-utils developers
#
# This file is part of python-litecoin-utils
#
# It is subject to the license terms in the LICENSE file found in the top-level
# directory of this distribution.
#
# No part of python-litecoin-utils, including this file, may be copied, modified,
# propagated, or distributed except according to the terms contained in the
# LICENSE file.

from decimal import Decimal

# TODO organise constants in sections

NETWORK_DEFAULT_PORTS = { 'mainnet': 9332,
                          'testnet': 19332,
                          'regtest': 19443 }

NETWORK_WIF_PREFIXES = { 'mainnet': b'\xB0',
                         'testnet': b'\xef',
                         'regtest': b'\xef' }

NETWORK_P2PKH_PREFIXES = { 'mainnet': b'\x30',
                           'testnet': b'\x6f',
                           'regtest': b'\x6f' }

NETWORK_P2SH_PREFIXES = { 'mainnet': b'\x05',
                          'testnet': b'\xc4',
                          'regtest': b'\xc4' }

NETWORK_SEGWIT_PREFIXES = { 'mainnet' : 'ltc',
                            'testnet' : 'tltc',
                            'regtest' : 'ltcrt' }

P2PKH_ADDRESS = "p2pkh"
P2SH_ADDRESS = "p2sh"
P2WPKH_ADDRESS_V0 = "p2wpkhv0"
P2WSH_ADDRESS_V0 = "p2wshv0"

SIGHASH_ALL = 0x01
SIGHASH_NONE = 0x02
SIGHASH_SINGLE = 0x03
SIGHASH_ANYONECANPAY = 0x80

TYPE_ABSOLUTE_TIMELOCK = 0x101
TYPE_RELATIVE_TIMELOCK = 0x201
TYPE_REPLACE_BY_FEE    = 0x301

DEFAULT_TX_LOCKTIME = b'\x00\x00\x00\x00'

EMPTY_TX_SEQUENCE = b'\x00\x00\x00\x00'
DEFAULT_TX_SEQUENCE = b'\xff\xff\xff\xff'
ABSOLUTE_TIMELOCK_SEQUENCE = b'\xfe\xff\xff\xff'

REPLACE_BY_FEE_SEQUENCE = b'\x01\x00\x00\x00'


# TX version 2 was introduced in BIP-68 with relative locktime -- tx v1
# does not support relative locktime
DEFAULT_TX_VERSION  = b'\x02\x00\x00\x00'

SATOSHIS_PER_BITCOIN = 100000000
NEGATIVE_SATOSHI = -1




# Copyright (C) 2018-2020 The python-litecoin-utils developers
#
# This file is part of python-litecoin-utils
#
# It is subject to the license terms in the LICENSE file found in the top-level
# directory of this distribution.
#
# No part of python-litecoin-utils, including this file, may be copied, modified,
# propagated, or distributed except according to the terms contained in the
# LICENSE file.

import math
import hashlib
import struct
from binascii import unhexlify, hexlify

from litecoinutils.constants import DEFAULT_TX_SEQUENCE, DEFAULT_TX_LOCKTIME, \
                    DEFAULT_TX_VERSION, NEGATIVE_SATOSHI, \
                    EMPTY_TX_SEQUENCE, SIGHASH_ALL, SIGHASH_NONE, \
                    SIGHASH_SINGLE, SIGHASH_ANYONECANPAY, \
                    ABSOLUTE_TIMELOCK_SEQUENCE, REPLACE_BY_FEE_SEQUENCE, \
                    TYPE_ABSOLUTE_TIMELOCK, TYPE_RELATIVE_TIMELOCK, \
                    TYPE_REPLACE_BY_FEE, SATOSHIS_PER_BITCOIN
from litecoinutils.script import Script
from litecoinutils.utils import to_bytes, vi_to_int

class TxInput:
    """Represents a transaction input.

    A transaction input requires a transaction id of a UTXO and the index of
    that UTXO.

    Attributes
    ----------
    txid : str
        the transaction id as a hex string (little-endian as displayed by
        tools)
    txout_index : int
        the index of the UTXO that we want to spend
    script_sig : list (strings)
        the op code and data of the script as string
    sequence : bytes
        the input sequence (for timelocks, RBF, etc.)

    Methods
    -------
    stream()
        converts TxInput to bytes
    copy()
        creates a copy of the object (classmethod)
    """

    def __init__(self, txid, txout_index, script_sig=Script([]), sequence=DEFAULT_TX_SEQUENCE):
        """See TxInput description"""

        # expected in the format used for displaying Bitcoin hashes
        self.txid = txid
        self.txout_index = txout_index
        self.script_sig = script_sig

        # if user provided a sequence it would be as string (for now...)
        if type(sequence) is str:
            self.sequence = unhexlify(sequence)
        else:
            self.sequence = sequence


    def stream(self):
        """Converts to bytes"""

        # Internally Bitcoin uses little-endian byte order as it improves
        # speed. Hashes are defined and implemented as big-endian thus
        # those are transmitted in big-endian order. However, when hashes are
        # displayed Bitcoin uses little-endian order because it is sometimes
        # convenient to consider hashes as little-endian integers (and not
        # strings)
        # - note that we reverse the byte order for the tx hash since the string
        #   was displayed in little-endian!
        # - note that python's struct uses little-endian by default
        txid_bytes = unhexlify(self.txid)[::-1]
        txout_bytes = struct.pack('<L', self.txout_index)
        script_sig_bytes = self.script_sig.to_bytes()
        data = txid_bytes + txout_bytes + \
                struct.pack('B', len(script_sig_bytes)) + \
                script_sig_bytes + self.sequence
        return data


    def __str__(self):
        return str({
            "txid": self.txid,
            "txout_index": self.txout_index,
            "script_sig": self.script_sig
        })

    def __repr__(self):
        return self.__str__()


    @staticmethod
    def from_raw(txinputraw, cursor=0, has_segwit=False):
        """
        Imports a TxInput from a Transaction's hexadecimal data

        Attributes
        ----------
        txinputraw : string (hex)
            The hexadecimal raw string of the Transaction
        cursor : int
            The cursor of which the algorithm will start to read the data
        has_segwit : boolean
            Is the Tx Input segwit or not
        """
        txinputraw = to_bytes(txinputraw)

        # read the 32 bytes of TxInput ID
        inp_hash = txinputraw[cursor:cursor + 32][::-1]

        if not len(inp_hash):
            raise Exception("Input transaction hash not found. Probably malformed raw transaction")
        output_n = txinputraw[cursor + 32:cursor + 36][::-1]
        cursor += 36

        # read the size (bytes length) of the integer representing the size of the Script's raw
        # data and the size of the Script's raw data
        unlocking_script_size, size = vi_to_int(txinputraw[cursor:cursor + 9])
        cursor += size
        unlocking_script = txinputraw[cursor:cursor + unlocking_script_size]
        cursor += unlocking_script_size
        sequence_number = txinputraw[cursor:cursor + 4]
        cursor += 4
        return TxInput(txid = inp_hash.hex(),
                       txout_index=int(output_n.hex(), 16),
                       script_sig=Script.from_raw(unlocking_script,has_segwit=has_segwit), sequence=sequence_number),cursor


    @classmethod
    def copy(cls, txin):
        """Deep copy of TxInput"""

        return cls(txin.txid, txin.txout_index, txin.script_sig,
                       txin.sequence)



class TxOutput:
    """Represents a transaction output

    Attributes
    ----------
    amount : int/float/Decimal
        the value we want to send to this output in satoshis
    script_pubkey : list (string)
        the script that will lock this amount

    Methods
    -------
    stream()
        converts TxInput to bytes
    copy()
        creates a copy of the object (classmethod)
    """


    def __init__(self, amount, script_pubkey):
        """See TxOutput description"""

        if not isinstance(amount, int):
            raise TypeError("Amount needs to be in satoshis as an integer")

        self.amount = amount
        self.script_pubkey = script_pubkey


    def stream(self):
        """Converts to bytes"""

        # internally all little-endian except hashes
        # note struct uses little-endian by default

        amount_bytes = struct.pack('<q', self.amount)
        script_bytes = self.script_pubkey.to_bytes()
        data = amount_bytes + struct.pack('B', len(script_bytes)) + script_bytes
        return data


    @staticmethod
    def from_raw(txoutputraw,cursor=0,has_segwit=False):
        """
        Imports a TxOutput from a Transaction's hexadecimal data

        Attributes
        ----------
        txinputraw : string (hex)
            The hexadecimal raw string of the Transaction
        cursor : int
            The cursor of which the algorithm will start to read the data
        has_segwit : boolean
            Is the Tx Output segwit or not
        """
        txoutputraw = to_bytes(txoutputraw)

        # read the amount of the TxOutput
        value = int.from_bytes(txoutputraw[cursor:cursor + 8][::-1], 'big')
        cursor += 8

        # read the size (bytes length) of the integer representing the size of the locking
        # Script's raw data and the size of the locking Script's raw data
        lock_script_size, size = vi_to_int(txoutputraw[cursor:cursor + 9])
        cursor += size
        lock_script = txoutputraw[cursor:cursor + lock_script_size]
        cursor += lock_script_size
        return TxOutput(amount=value,
                        script_pubkey=Script.from_raw(lock_script, has_segwit=has_segwit)),cursor



    def __str__(self):
        return str({
            "amount": self.amount,
            "script_pubkey": self.script_pubkey
        })

    def __repr__(self):
        return self.__str__()


    @classmethod
    def copy(cls, txout):
        """Deep copy of TxOutput"""

        return cls(txout.amount, txout.script_pubkey)


class Sequence:
    """Helps setting up appropriate sequence. Used to provide the sequence to
    transaction inputs and to scripts.

    Attributes
    ----------
    value : int
        The value of the block height or the 512 seconds increments
    seq_type : int
        Specifies the type of sequence (TYPE_RELATIVE_TIMELOCK |
        TYPE_ABSOLUTE_TIMELOCK | TYPE_REPLACE_BY_FEE
    is_type_block : bool
        If type is TYPE_RELATIVE_TIMELOCK then this specifies its type
        (block height or 512 secs increments)

    Methods
    -------
    for_input_sequence()
        Serializes the relative sequence as required in a transaction
    for_script()
        Returns the appropriate integer for a script; e.g. for relative timelocks

    Raises
    ------
    ValueError
        if the value is not within range of 2 bytes.
    """

    def __init__(self, seq_type, value=None, is_type_block=True):
        self.seq_type = seq_type
        self.value = value
        if self.seq_type == TYPE_RELATIVE_TIMELOCK and (self.value < 1 or self.value > 0xffff):
            raise ValueError('Sequence should be between 1 and 65535')
        self.is_type_block = is_type_block

    def for_input_sequence(self):
        """Creates a relative timelock sequence value as expected from
        TxInput sequence attribute"""
        if self.seq_type == TYPE_ABSOLUTE_TIMELOCK:
            return ABSOLUTE_TIMELOCK_SEQUENCE

        if self.seq_type == TYPE_REPLACE_BY_FEE:
            return REPLACE_BY_FEE_SEQUENCE

        if self.seq_type == TYPE_RELATIVE_TIMELOCK:
            # most significant bit is already 0 so relative timelocks are enabled
            seq = 0
            # if not block height type set 23 bit
            if not self.is_type_block:
                seq |= 1 << 22
            # set the value
            seq |= self.value
            seq_bytes = seq.to_bytes(4, byteorder='little')
            return seq_bytes



    def for_script(self):
        """Creates a relative/absolute timelock sequence value as expected in scripts"""
        if self.seq_type == TYPE_REPLACE_BY_FEE:
            raise ValueError('RBF is not to be included in a script.')

        script_integer = self.value

        # if not block-height type then set 23 bit
        if self.seq_type == TYPE_RELATIVE_TIMELOCK and not self.is_type_block:
            script_integer |= 1 << 22

        return script_integer


class Locktime:
    """Helps setting up appropriate locktime.

    Attributes
    ----------
    value : int
        The value of the block height or the Unix epoch (seconds from 1 Jan
        1970 UTC)

    Methods
    -------
    for_transaction()
        Serializes the locktime as required in a transaction

    Raises
    ------
    ValueError
        if the value is not within range of 2 bytes.
    """

    def __init__(self, value):
        self.value = value

    def for_transaction(self):
        """Creates a timelock as expected from Transaction"""

        locktime_bytes = self.value.to_bytes(4, byteorder='little')
        return locktime_bytes



class Transaction:
    """Represents a Bitcoin transaction

    Attributes
    ----------
    inputs : list (TxInput)
        A list of all the transaction inputs
    outputs : list (TxOutput)
        A list of all the transaction outputs
    locktime : bytes
        The transaction's locktime parameter
    version : bytes
        The transaction version
    has_segwit : bool
        Specifies a tx that includes segwit inputs
    witnesses : list (Script)
        The witness scripts that correspond to the inputs


    Methods
    -------
    stream()
        Converts Transaction to bytes
    serialize()
        Converts Transaction to hex string
    get_txid()
        Calculates txid and returns it
    get_hash()
        Calculates tx hash (wtxid) and returns it
    get_wtxid()
        Calculates tx hash (wtxid) and returns it
    get_size()
        Calculates the tx size
    get_vsize()
        Calculates the tx segwit size
    copy()
        creates a copy of the object (classmethod)
    get_transaction_digest(txin_index, script, sighash)
        returns the transaction input's digest that is to be signed according
    get_transaction_segwit_digest(txin_index, script, amount, sighash)
        returns the transaction input's segwit digest that is to be signed
        according to sighash
    """

    def __init__(self, inputs=None, outputs=None, locktime=DEFAULT_TX_LOCKTIME,
                 version=DEFAULT_TX_VERSION, has_segwit=False, witnesses=None):
        """See Transaction description"""

        # make sure default argument for inputs, outputs and witnesses is an empty list
        if inputs is None:
            inputs = []
        if outputs is None:
            outputs = []
        if witnesses is None:
            witnesses = []

        self.inputs = inputs
        self.outputs = outputs
        self.has_segwit = has_segwit
        self.witnesses = witnesses

        # if user provided a locktime it would be as string (for now...)
        if type(locktime) is str:
            self.locktime = unhexlify(locktime)
        else:
            self.locktime = locktime

        self.version = version


    @staticmethod
    def from_raw(txraw):
        """
        Imports a Transaction from hexadecimal data

        Attributes
        ----------
        txinputraw : string (hex)
            The hexadecimal raw string of the Transaction
        cursor : int
            The cursor of which the algorithm will start to read the data
        has_segwit : boolean
            Is the Tx Input segwit or not
        """
        rawtx = to_bytes(txraw)

        # read version
        version = rawtx[0:4][::-1]
        flag = None
        has_segwit = False
        cursor = 4
        if rawtx[4:5] == b'\0':
            flag = rawtx[5:6]
            if flag == b'\1':
                has_segwit = True
            cursor += 2

        # read the size (bytes length) of the integer representing the size of the inputs
        # number and the the inputs number

        n_inputs, size = vi_to_int(rawtx[cursor:cursor + 9])
        cursor += size
        inputs = []

        #iterate n_inputs times to read the inputs from raw
        for index in range(0,n_inputs):
            inp, cursor = TxInput.from_raw(rawtx, cursor=cursor, has_segwit=has_segwit)
            inputs.append(inp)

        outputs = []
        # read the size (bytes length) of the integer representing the size of the outputs
        # number and the the outputs number
        n_outputs, size = vi_to_int(rawtx[cursor:cursor + 9])
        cursor += size
        output_total = 0

        # iterate n_outputs times to read the inputs from raw
        for index in range(0,n_outputs):
            output, cursor = TxOutput.from_raw(rawtx, cursor=cursor, has_segwit=has_segwit)
            outputs.append(output)

        witnesses = []
        if has_segwit == True:
            # iterate to read the witnesses for every input
            for n in range(0, len(inputs)):
                n_items, size = vi_to_int(rawtx[cursor:cursor + 9])
                cursor += size
                witnesses_tmp = []
                for m in range(0, n_items):
                    witness = b'\0'
                    item_size, size = vi_to_int(rawtx[cursor:cursor + 9])
                    if item_size:
                        witness = rawtx[cursor + size:cursor + item_size + size]
                    cursor += item_size + size
                    witnesses_tmp.append(witness.hex())
                witnesses.append(Script(script=witnesses_tmp))

        return Transaction(inputs = inputs,
                           outputs = outputs,
                           has_segwit = has_segwit,
                           witnesses = witnesses)



    def __str__(self):
        return str({
                "inputs": self.inputs,
                "outputs": self.outputs,
                "has_segwit": self.has_segwit,
                "witnesses": self.witnesses,
                "locktime": self.locktime.hex(),
                "version": self.version.hex()
                })

    def __repr__(self):
        return self.__str__()


    @classmethod
    def copy(cls, tx):
        """Deep copy of Transaction"""

        ins = [TxInput.copy(txin) for txin in tx.inputs]
        outs = [TxOutput.copy(txout) for txout in tx.outputs]
        wits = [Script.copy(witness) for witness in tx.witnesses]
        return cls(ins, outs, tx.locktime, tx.version, tx.has_segwit, wits)


    def get_transaction_digest(self, txin_index, script, sighash=SIGHASH_ALL):
        """Returns the transaction's digest for signing.

        |  SIGHASH types (see constants.py):
        |      SIGHASH_ALL - signs all inputs and outputs (default)
        |      SIGHASH_NONE - signs all of the inputs
        |      SIGHASH_SINGLE - signs all inputs but only txin_index output
        |      SIGHASH_ANYONECANPAY (only combined with one of the above)
        |      - with ALL - signs all outputs but only txin_index input
        |      - with NONE - signs only the txin_index input
        |      - with SINGLE - signs txin_index input and output

        Attributes
        ----------
        txin_index : int
            The index of the input that we wish to sign
        script : list (string)
            The scriptPubKey of the UTXO that we want to spend
        sighash : int
            The type of the signature hash to be created
        """

        # clone transaction to modify without messing up the real transaction
        tmp_tx = Transaction.copy(self)

        # make sure all input scriptSigs are empty
        for txin in tmp_tx.inputs:
            txin.script_sig = Script([])

        #
        # TODO Deal with (delete?) script's OP_CODESEPARATORs, if any
        # Very early versions of Bitcoin were using a different design for
        # scripts that were flawed. OP_CODESEPARATOR has no purpose currently
        # but we could not delete it for compatibility purposes. If it exists
        # in a script it needs to be removed.
        #

        # the temporary transaction's scriptSig needs to be set to the
        # scriptPubKey of the UTXO we are trying to spend - this is required to
        # get the correct transaction digest (which is then signed)
        tmp_tx.inputs[txin_index].script_sig = script

        #
        # by default we sign all inputs/outputs (SIGHASH_ALL is used)
        #

        # whether 0x0n or 0x8n, bitwise AND'ing will result to n
        if (sighash & 0x1f) == SIGHASH_NONE:
            # do not include outputs in digest (i.e. do not sign outputs)
            tmp_tx.outputs = []

            # do not include sequence of other inputs (zero them for digest)
            # which means that they can be replaced
            for i in range(len(tmp_tx.inputs)):
                if i != txin_index:
                    tmp_tx.inputs[i].sequence = EMPTY_TX_SEQUENCE

        elif (sighash & 0x1f) == SIGHASH_SINGLE:
            # only sign the output that corresponds to txin_index

            if txin_index >= len(tmp_tx.outputs):
                raise ValueError('Transaction index is greater than the \
                                 available outputs')

            # keep only output that corresponds to txin_index -- delete all outputs
            # after txin_index and zero out all outputs upto txin_index
            txout = tmp_tx.outputs[txin_index]
            tmp_tx.outputs = []
            for i in range(txin_index):
                tmp_tx.outputs.append( TxOutput(NEGATIVE_SATOSHI, Script([])) )
            tmp_tx.outputs.append(txout)

            # do not include sequence of other inputs (zero them for digest)
            # which means that they can be replaced
            for i in range(len(tmp_tx.inputs)):
                if i != txin_index:
                    tmp_tx.inputs[i].sequence = EMPTY_TX_SEQUENCE

        # bitwise AND'ing 0x8n to 0x80 will result to true
        if sighash & SIGHASH_ANYONECANPAY:
            # ignore all other inputs from the signature which means that
            # anyone can add new inputs
            tmp_tx.inputs = [tmp_tx.inputs[txin_index]]

        # get the byte stream of the temporary transaction
        tx_for_signing = tmp_tx.stream(False)

        # add sighash bytes to be hashed
        # Note that although sighash is one byte it is hashed as a 4 byte value.
        # There is no real reason for this other than that the original implementation
        # of Bitcoin stored sighash as an integer (which serializes as a 4
        # bytes), i.e. it should be converted to one byte before serialization.
        # It is converted to 1 byte before serializing to send to the network
        tx_for_signing += struct.pack('<i', sighash)

        # create transaction digest -- note double hashing
        tx_digest = hashlib.sha256( hashlib.sha256(tx_for_signing).digest()).digest()

        return tx_digest


    def get_transaction_segwit_digest(self, txin_index, script, amount, sighash=SIGHASH_ALL):
        """Returns the segwit transaction's digest for signing.

                |  SIGHASH types (see constants.py):
                |      SIGHASH_ALL - signs all inputs and outputs (default)
                |      SIGHASH_NONE - signs all of the inputs
                |      SIGHASH_SINGLE - signs all inputs but only txin_index output
                |      SIGHASH_ANYONECANPAY (only combined with one of the above)
                |      - with ALL - signs all outputs but only txin_index input
                |      - with NONE - signs only the txin_index input
                |      - with SINGLE - signs txin_index input and output

                Attributes
                ----------
                txin_index : int
                    The index of the input that we wish to sign
                script : list (string)
                    The scriptPubKey of the UTXO that we want to spend
                amount : int/float/Decimal
                    The amount of the UTXO to spend is included in the
                    signature for segwit (in satoshis)
                sighash : int
                    The type of the signature hash to be created
                """

        # clone transaction to modify without messing up the real transaction
        tmp_tx = Transaction.copy(self)

        # TODO consult ref. impl. in BIP-143 and update if needed
        # requires cleanup and further explanations
        hash_prevouts = b'\x00' * 32
        hash_sequence = b'\x00' * 32
        hash_outputs = b'\x00' * 32

        # Judging the signature type
        basic_sig_hash_type = (sighash & 0x1f)
        anyone_can_pay = sighash& 0xf0 == SIGHASH_ANYONECANPAY
        sign_all = (basic_sig_hash_type != SIGHASH_SINGLE) and (basic_sig_hash_type != SIGHASH_NONE)

        # Hash all input
        if not anyone_can_pay:
            hash_prevouts = b''
            for txin in tmp_tx.inputs:
                hash_prevouts += unhexlify(txin.txid)[::-1] + \
                                    struct.pack('<L', txin.txout_index)
            hash_prevouts = hashlib.sha256(hashlib.sha256(hash_prevouts).digest()).digest()

        # Hash all input sequence
        if not anyone_can_pay and sign_all:
            hash_sequence = b''
            for txin in tmp_tx.inputs:
                hash_sequence += txin.sequence
            hash_sequence = hashlib.sha256(hashlib.sha256(hash_sequence).digest()).digest()

        if sign_all:
            # Hash all output
            hash_outputs = b''
            for txout in tmp_tx.outputs:
                amount_bytes = struct.pack('<q', txout.amount)
                script_bytes = txout.script_pubkey.to_bytes()
                hash_outputs += amount_bytes + struct.pack('B', len(script_bytes)) + script_bytes
            hash_outputs = hashlib.sha256(hashlib.sha256(hash_outputs).digest()).digest()
        elif basic_sig_hash_type == SIGHASH_SINGLE and txin_index < len(tmp_tx.outputs):
            # Hash one output
            txout = tmp_tx.outputs[txin_index]
            amount_bytes = struct.pack('<q', txout.amount)
            script_bytes = txout.script_pubkey.to_bytes()
            hash_outputs = amount_bytes + struct.pack('B', len(script_bytes)) + script_bytes
            hash_outputs = hashlib.sha256(hashlib.sha256(hash_outputs).digest()).digest()

        # add sighash bytes to be hashed
        tx_for_signing = self.version

        # add sighash bytes to be hashed
        tx_for_signing += hash_prevouts + hash_sequence

        # add tx txin
        txin = self.inputs[txin_index]
        tx_for_signing += unhexlify(txin.txid)[::-1] + \
                          struct.pack('<L', txin.txout_index)

        # add tx sign
        tx_for_signing += struct.pack('B', len(script.to_bytes()))
        tx_for_signing += script.to_bytes()

        # add txin amount
        tx_for_signing += struct.pack('<q', amount)

        # add tx sequence
        tx_for_signing += txin.sequence

        # add txouts hash
        tx_for_signing += hash_outputs

        # add locktime
        tx_for_signing += self.locktime

        # add sighash type
        tx_for_signing += struct.pack('<i', sighash)

        return hashlib.sha256(hashlib.sha256(tx_for_signing).digest()).digest()


    def stream(self, has_segwit):
        """Converts to bytes"""

        data = self.version
        if has_segwit and self.witnesses:
            # marker
            data += b'\x00'
            # flag
            data += b'\x01'

        txin_count_bytes = chr(len(self.inputs)).encode()
        txout_count_bytes = chr(len(self.outputs)).encode()
        data += txin_count_bytes
        for txin in self.inputs:
            data += txin.stream()
        data += txout_count_bytes
        for txout in self.outputs:
            data += txout.stream()
        if has_segwit:
            for witness in self.witnesses:
                # add witnesses script Count
                witnesses_count_bytes = chr(len(witness.script)).encode()
                data += witnesses_count_bytes
                data += witness.to_bytes(True)
        data += self.locktime
        return data


    def get_txid(self):
        """Hashes the serialized (bytes) tx to get a unique id"""

        data = self.stream(False)
        hash = hashlib.sha256( hashlib.sha256(data).digest() ).digest()
        # note that we reverse the hash for display purposes
        return hexlify(hash[::-1]).decode('utf-8')


    def get_wtxid(self):
        """Hashes the serialized (bytes) tx including segwit marker and witnesses"""

        return get_hash()


    def get_hash(self):
        """Hashes the serialized (bytes) tx including segwit marker and witnesses"""

        data = self.stream(self.has_segwit)
        hash = hashlib.sha256( hashlib.sha256(data).digest() ).digest()
        # note that we reverse the hash for display purposes
        return hexlify(hash[::-1]).decode('utf-8')


    def get_size(self):
        """Gets the size of the transaction"""

        return len(self.stream(self.has_segwit))


    def get_vsize(self):
        """Gets the virtual size of the transaction.

        For non-segwit txs this is identical to get_size(). For segwit txs the
        marker and witnesses length needs to be reduced to 1/4 of its original
        length. Thus it is substructed from size and then it is divided by 4
        before added back to size to produce vsize (always rounded up).

        https://en.litecoin.it/wiki/Weight_units
        """
        # return size if non segwit
        if not self.has_segwit:
            return self.get_size()

        marker_size = 2

        wit_size = 0
        data = b''

        # count witnesses data
        for witness in self.witnesses:
            # add witnesses script Count
            witnesses_count_bytes = chr(len(witness.script)).encode()
            data = witnesses_count_bytes
            data += witness.to_bytes(True)
        wit_size = len(data)
        # TODO when TxInputWitness is created it will contain it's own len or
        # size method

        size = self.get_size() - (marker_size + wit_size)
        vsize = size + (marker_size + wit_size) / 4

        return int( math.ceil(vsize) )


    def serialize(self):
        """Converts to hex string"""

        return hexlify(self.stream(self.has_segwit)).decode('utf-8')


def main():
    pass

if __name__ == "__main__":
    main()




from setuptools import setup
from litecoinutils import __version__

#with open('requirements.txt') as f:
#    requirements = f.read().splitlines()

#install_reqs = parse_requirements('requirements.txt', session=False)
#requirements = [str(ir.req) for ir in install_reqs]

with open('README.rst') as readme:
    long_description = readme.read()

setup(name='litecoin-utils',
      version=__version__,
      description='Litecoin utility functions',
      long_description=long_description,
      author='Konstantinos Karasavvas',
      author_email='kkarasavvas@gmail.com',
      url='https://github.com/karask/python-litecoin-utils',
      license='AGPLv3',
      keywords='litecoin library utilities tools',
      install_requires=[
          'base58check==1.0.2',
          'ecdsa==0.17.0',
          'sympy==1.3',
          'python-bitcoinrpc==1.0'
      ],
      packages=['litecoinutils'],
      #package_data={
      #    'litecoinutils': ['requirements.txt']
      #},
      #include_package_data=True,
      zip_safe=False
     )




# Copyright (C) 2018-2020 The python-litecoin-utils developers
#
# This file is part of python-litecoin-utils
#
# It is subject to the license terms in the LICENSE file found in the top-level
# directory of this distribution.
#
# No part of python-litecoin-utils, including this file, may be copied, modified,
# propagated, or distributed except according to the terms contained in the
# LICENSE file.

from binascii import hexlify, unhexlify

from litecoinutils.constants import SATOSHIS_PER_BITCOIN


def to_satoshis(num):
    '''
    Converts from any number type (int/float/Decimal) to satoshis (int)
    '''

    # we need to round because of how floats are stored insternally:
    # e.g. 0.29 * 100000000 = 28999999.999999996
    return int( round(num * SATOSHIS_PER_BITCOIN) )

def encode_varint(i):
    '''
    Encode a potentially very large integer into varint bytes. The length should be
    specified in little-endian.
    https://bitcoin.org/en/developer-reference#compactsize-unsigned-integers
    '''
    if i < 253:
        return bytes([i])
    elif i < 0x10000:
        return b'\xfd' +  i.to_bytes(2, 'little')
    elif i < 0x100000000:
        return b'\xfe' +  i.to_bytes(4, 'little')
    elif i < 0x10000000000000000:
        return b'\xff' +  i.to_bytes(8, 'little')
    else:
        raise ValueError("Integer is too large: %d" % i)

def prepend_compact_size(data):
    '''
    Counts bytes and returns them with their compact size (or varint) prepended.
    Accepts bytes and returns bytes. The length should be specified in
    little-endian (which is why we reverse the array bytes).

    https://bitcoin.org/en/developer-reference#compactsize-unsigned-integers
    '''

    prefix = b''
    size = len(data)
    if size >= 0 and size <= 252:
        prefix = unhexlify(format(size, '02x').encode())
    elif size >= 253 and size <= 0xffff:
        prefix = b'\xfd' + unhexlify(format(size, '04x'))[::-1]
    elif size >= 0x10000 and size <= 0xffffffff:
        prefix = b'\xfe' + unhexlify(format(size, '08x'))[::-1]
    elif size >= 0x100000000 and size <= 0xffffffffffffffff:
        prefix = b'\xff' + unhexlify(format(size, '016x'))[::-1]
    else:
        raise ValueError("Data size not between 0 and 0xffffffffffffffff")

    return prefix + data


def is_address_bech32(address):
    '''
    Returns if an address (string) is bech32 or not
    TODO improve by checking if valid, etc.
    '''
    if (address.startswith('ltc') or
        address.startswith('tltc')):
        return True

    return False



def vi_to_int(byteint):
    '''
    Converts var_int (bytes) to int
    '''
    if not isinstance(byteint, (bytes)):
        raise Exception("Byteint must be a list or defined as bytes")
    ni = byteint[0]
    if ni < 253:
        return ni, 1
    if ni == 253:  # integer of 2 bytes
        size = 2
    elif ni == 254:  # integer of 4 bytes
        size = 4
    else:  # integer of 8 bytes
        size = 8
    return int.from_bytes(byteint[1:1+size][::-1], 'big'), size + 1


def to_bytes(string, unhexlify=True):
    '''
	Converts a hex string to bytes
    '''
    if not string:
        return b''
    if unhexlify:
        try:
            if isinstance(string, bytes):
                string = string.decode()
            s = bytes.fromhex(string)
            return s
        except (TypeError, ValueError):
            pass
    if isinstance(string, bytes):
        return string
    else:
        return bytes(string, 'utf8')


